/*  XRoar - a Dragon/Tandy Coco emulator
 *  Copyright (C) 2003-2011  Ciaran Anscomb
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

#include <stdlib.h>
#include <string.h>
#include <SDL.h>
#ifdef GEKKO
#include <SDL/SDL_ttf.h>
#include "gctypes.h"
#include "dejavusans_ttf.h"
#include "another_ttf.h"
#endif
#include <SDL/SDL_syswm.h>
#include "types.h"
#include "logging.h"
#include "module.h"
#include "vdg_bitmaps.h"
#include "xroar.h"
#ifdef WINDOWS32
#include "windows32/common_windows32.h"
#endif

static int init(void);
static void shutdown(void);
static void alloc_colours(void);
static void reset(void);
static void vsync(void);
static void hsync(void);
static void set_mode(unsigned int mode);
static void render_border(void);
static int set_fullscreen(int fullscreen);

VideoModule video_sdl_module = {
	.common = { .name = "sdl", .description = "Standard SDL surface",
	            .init = init, .shutdown = shutdown },
	.update_palette = alloc_colours,
	.reset = reset, .vsync = vsync, .hsync = hsync, .set_mode = set_mode,
	.render_border = render_border, .set_fullscreen = set_fullscreen,
};

typedef Uint32 Pixel;
#define RESET_PALETTE() reset_palette()
#define MAPCOLOUR(r,g,b) alloc_and_map(r, g, b)
#define VIDEO_SCREENBASE ((Pixel *)screen->pixels)
#define XSTEP 1
#define NEXTLINE 0
#define VIDEO_TOPLEFT (VIDEO_SCREENBASE)
#define VIDEO_VIEWPORT_YOFFSET (0)
#define LOCK_SURFACE SDL_LockSurface(screen)
#define UNLOCK_SURFACE SDL_UnlockSurface(screen)
#define VIDEO_MODULE_NAME video_sdl_module

#ifdef GEKKO
TTF_Font *text_fontSmall, *text_fontLarge;
SDL_Surface  *openDialog, *screen;
#endif

static int palette_index = 0;

static void reset_palette(void) {
	palette_index = 0;
}

static Pixel alloc_and_map(int r, int g, int b) {
	SDL_Color c;
	c.r = r;
	c.g = g;
	c.b = b;
	SDL_SetPalette(screen, SDL_LOGPAL|SDL_PHYSPAL, &c, palette_index, 1);
	palette_index++;
	palette_index %= 256;
	return SDL_MapRGB(screen->format, r, g, b);
}

#include "vo_generic_ops.h"

static int init(void) {
	LOG_DEBUG(2,"Initialising SDL video driver\n");
#ifdef WINDOWS32
	if (!getenv("SDL_VIDEODRIVER"))
		putenv("SDL_VIDEODRIVER=windib");
#endif

#ifdef GEKKO
	char *openDialogfname;
	openDialog = SDL_CreateRGBSurface(0, 320, 240, 32, 0, 0, 0, 0);

	/* init SDL TTF */
	TTF_Init();

	SDL_RWops* fontsmallRW = SDL_RWFromMem((char *)another_ttf, another_ttf_size);
	text_fontSmall = TTF_OpenFontRW(fontsmallRW, 1, 12); 
	SDL_RWops* fontlargeRW = SDL_RWFromMem((char *)dejavusans_ttf, dejavusans_ttf_size);
	text_fontLarge = TTF_OpenFontRW(fontlargeRW, 1, 13); 
#endif
	if (!SDL_WasInit(SDL_INIT_NOPARACHUTE)) {
		if (SDL_Init(SDL_INIT_NOPARACHUTE) < 0) {
			LOG_ERROR("Failed to initialise SDL: %s\n", SDL_GetError());
			return 1;
		}
	}
	if (SDL_InitSubSystem(SDL_INIT_VIDEO) < 0) {
		LOG_ERROR("Failed to initialise SDL video driver: %s\n", SDL_GetError());
		return 1;
	}
	if (set_fullscreen(xroar_opt_fullscreen))
		return 1;
#ifdef WINDOWS32
	{
		SDL_version sdlver;
		SDL_SysWMinfo sdlinfo;
		SDL_VERSION(&sdlver);
		sdlinfo.version = sdlver;
		SDL_GetWMInfo(&sdlinfo);
		windows32_main_hwnd = sdlinfo.window;
	}
#endif
	reset();
	set_mode(0);
	return 0;
}

static void shutdown(void) {
	LOG_DEBUG(2,"Shutting down SDL video driver\n");
	set_fullscreen(0);
	/* Should not be freed by caller: SDL_FreeSurface(screen); */
	SDL_QuitSubSystem(SDL_INIT_VIDEO);
}

static int set_fullscreen(int fullscreen) {
	screen = SDL_SetVideoMode(320, 240, 32, SDL_HWSURFACE|(fullscreen?SDL_FULLSCREEN:0));
	SDL_BlitSurface(openDialog, 0/*SDL_Rect *srcrect*/, screen, 0/*SDL_Rect *dstrect*/);
	if (screen == NULL) {
		LOG_ERROR("Failed to allocate SDL surface for display\n");
		return 1;
	}
	pixel = VIDEO_TOPLEFT + VIDEO_VIEWPORT_YOFFSET;
	alloc_colours();
	if (fullscreen)
		SDL_ShowCursor(SDL_DISABLE);
	else
		SDL_ShowCursor(SDL_ENABLE);
	video_sdl_module.is_fullscreen = fullscreen;
#ifdef GEKKO
	SDL_ShowCursor(SDL_DISABLE);
#endif
	return 0;
}

static void reset(void) {
	pixel = VIDEO_TOPLEFT + VIDEO_VIEWPORT_YOFFSET;
	subline = 0;
	beam_pos = 0;
}
static void vsync(void) {
	SDL_UpdateRect(screen, 0, 0, 320, 240);
	reset();
}
